@@ -1,3 +1,7 @@
+#include "request.h"
+#include "keyvalue.h"
+#include "log.h"
+
 #include <sys/stat.h>
 
 #include <limits.h>
@@ -6,10 +10,6 @@
 #include <stdio.h>
 #include <ctype.h>
 
-#include "request.h"
-#include "keyvalue.h"
-#include "log.h"
-
 static int request_check_hostname(server *srv, connection *con, buffer *host) {
 	enum { DOMAINLABEL, TOPLABEL } stage = TOPLABEL;
 	size_t i;
@@ -86,10 +86,18 @@ static int request_check_hostname(server *srv, connection *con, buffer *host) {
 	if (host_len == 0) return -1;
 
 	/* if the hostname ends in a "." strip it */
-	if (host->ptr[host_len-1] == '.') host_len -= 1;
+	if (host->ptr[host_len-1] == '.') {
+		/* shift port info one left */
+		if (NULL != colon) memmove(colon-1, colon, host->used - host_len);
+		else host->ptr[host_len-1] = '\0';
+		host_len -= 1;
+		host->used -= 1;
+	}
+
+	if (host_len == 0) return -1;
 
 	/* scan from the right and skip the \0 */
-	for (i = host_len - 1; i + 1 > 0; i--) {
+	for (i = host_len; i-- > 0; ) {
 		const char c = host->ptr[i];
 
 		switch (stage) {
@@ -103,7 +111,7 @@ static int request_check_hostname(server *srv, connection *con, buffer *host) {
 
 					/* check the first character at right of the dot */
 					if (is_ip == 0) {
-						if (!light_isalpha(host->ptr[i+1])) {
+						if (!light_isalnum(host->ptr[i+1])) {
 							return -1;
 						}
 					} else if (!light_isdigit(host->ptr[i+1])) {
@@ -125,7 +133,7 @@ static int request_check_hostname(server *srv, connection *con, buffer *host) {
 				}
 			} else if (i == 0) {
 				/* the first character of the hostname */
-				if (!light_isalpha(c)) {
+				if (!light_isalnum(c)) {
 					return -1;
 				}
 				label_len++;
@@ -200,7 +208,7 @@ static int request_check_hostname(server *srv, connection *con, buffer *host) {
 #define DUMP_HEADER
 #endif
 
-int http_request_split_value(array *vals, buffer *b) {
+static int http_request_split_value(array *vals, buffer *b) {
 	char *s;
 	size_t i;
 	int state = 0;
@@ -262,7 +270,7 @@ int http_request_split_value(array *vals, buffer *b) {
 	return 0;
 }
 
-int request_uri_is_valid_char(unsigned char c) {
+static int request_uri_is_valid_char(unsigned char c) {
 	if (c <= 32) return 0;
 	if (c == 127) return 0;
 	if (c == 255) return 0;
@@ -274,6 +282,8 @@ int http_request_parse(server *srv, connection *con) {
 	char *uri = NULL, *proto = NULL, *method = NULL, con_length_set;
 	int is_key = 1, key_len = 0, is_ws_after_key = 0, in_folding;
 	char *value = NULL, *key = NULL;
+	char *reqline_host = NULL;
+	int reqline_hostlen = 0;
 
 	enum { HTTP_CONNECTION_UNSET, HTTP_CONNECTION_KEEPALIVE, HTTP_CONNECTION_CLOSE } keep_alive_set = HTTP_CONNECTION_UNSET;
 
@@ -443,7 +453,14 @@ int http_request_parse(server *srv, connection *con) {
 
 				if (0 == strncmp(uri, "http://", 7) &&
 				    NULL != (nuri = strchr(uri + 7, '/'))) {
-					/* ignore the host-part */
+					reqline_host = uri + 7;
+					reqline_hostlen = nuri - reqline_host;
+
+					buffer_copy_string_len(con->request.uri, nuri, proto - nuri - 1);
+				} else if (0 == strncmp(uri, "https://", 8) &&
+				    NULL != (nuri = strchr(uri + 8, '/'))) {
+					reqline_host = uri + 8;
+					reqline_hostlen = nuri - reqline_host;
 
 					buffer_copy_string_len(con->request.uri, nuri, proto - nuri - 1);
 				} else {
@@ -532,8 +549,8 @@ int http_request_parse(server *srv, connection *con) {
 		con->response.keep_alive = 0;
 		con->keep_alive = 0;
 
-		log_error_write(srv, __FILE__, __LINE__, "s", "no uri specified -> 400");
 		if (srv->srvconf.log_request_header_on_error) {
+			log_error_write(srv, __FILE__, __LINE__, "s", "no uri specified -> 400");
 			log_error_write(srv, __FILE__, __LINE__, "Sb",
 							"request-header:\n",
 							con->request.request);
@@ -541,6 +558,19 @@ int http_request_parse(server *srv, connection *con) {
 		return 0;
 	}
 
+	if (reqline_host) {
+		/* Insert as host header */
+		data_string *ds;
+
+		if (NULL == (ds = (data_string *)array_get_unused_element(con->request.headers, TYPE_STRING))) {
+			ds = data_string_init();
+		}
+
+		buffer_copy_string_len(ds->key, CONST_STR_LEN("Host"));
+		buffer_copy_string_len(ds->value, reqline_host, reqline_hostlen);
+		array_insert_unique(con->request.headers, (data_unset *)ds);
+		con->request.http_host = ds->value;
+	}
 
 	for (; i < con->parse_request->used && !done; i++) {
 		char *cur = con->parse_request->ptr + i;
@@ -901,7 +931,11 @@ int http_request_parse(server *srv, connection *con) {
 									return 0;
 								}
 							} else if (cmp > 0 && 0 == (cmp = buffer_caseless_compare(CONST_BUF_LEN(ds->key), CONST_STR_LEN("Host")))) {
-								if (!con->request.http_host) {
+								if (reqline_host) {
+									/* ignore all host: headers as we got the host in the request line */
+									ds->free((data_unset*) ds);
+									ds = NULL;
+								} else if (!con->request.http_host) {
 									con->request.http_host = ds->value;
 								} else {
 									con->http_status = 400;
@@ -948,18 +982,8 @@ int http_request_parse(server *srv, connection *con) {
 								if (!con->request.http_if_none_match) {
 									con->request.http_if_none_match = ds->value->ptr;
 								} else {
-									con->http_status = 400;
-									con->keep_alive = 0;
-
-									if (srv->srvconf.log_request_header_on_error) {
-										log_error_write(srv, __FILE__, __LINE__, "s",
-												"duplicate If-None-Match-header -> 400");
-										log_error_write(srv, __FILE__, __LINE__, "Sb",
-												"request-header:\n",
-												con->request.request);
-									}
-									array_insert_unique(con->request.headers, (data_unset *)ds);
-									return 0;
+									ds->free((data_unset*) ds);
+									ds = NULL;
 								}
 							} else if (cmp > 0 && 0 == (cmp = buffer_caseless_compare(CONST_BUF_LEN(ds->key), CONST_STR_LEN("Range")))) {
 								if (!con->request.http_range) {
